#!/usr/bin/python
#   coding: utf-8

import os, re, time, pickle, sys
import xml.etree.cElementTree as etree
import pad.tree
import pad.helper
from pad.logger import Logger
logger = Logger()

def dateCode():
  return re.sub('[^0-9]+','',datetime.datetime.now().isoformat())

def nsquote(s):
  quoted = re.sub('([^0-9a-zA-Z .{}:/-])',
    lambda match:'['+hex(ord(match.group(1))+65536)[-4:]+']', s)
  quoted = quoted.replace(' ','_').replace(':',',')
  return quoted

def filenamequote(s):
  return nsquote(s).replace('/','+')

def filenameunquote(s):
  s = s.replace('+','/').replace('_',' ').replace(',',':')
  return re.sub('\[([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])\]',
    lambda match:unichr(int(match.group(1),16)), unicode(s))





class Delta:
  def __init__( self, location, version, conflict=[] ):
    self.location = location
    self.version = version
    self.conflict = conflict

  def swap( self, old ):
    '''Given a proposed delta (old) to the tree just before self,
returns a delta which can be applied just after self to the same effect.

Has no effect on the delta's version number. You must set the version number
at the end of the swap chain.'''

    return self

  def apply( self, tree ):
    '''Applies the delta to the tree supplied, in place (returns None)'''
    return None

  def __repr__( self ):
    return '<No change>'



class Delete(Delta):
  def __init__( self, location, version, oldTag, oldContent, conflict=[] ):
    if oldTag is None: oldTag=''
    if oldContent is None: oldContent=''

    Delta.__init__(self, location, version, conflict)
    self.oldContent = oldContent
    self.oldTag = oldTag

  def swap( self, old ):
    if old.location == self.location:
      if isinstance( old, Delete ):
        # new doesn't need to do anything anymore
        return Delta( self.location, self.version )
      elif isinstance( old, Replace ):
        # new must re-add the deleted element
        # this is really a conflict but no information is lost so it's safe to save and let the user fix later
        return AddBefore( self.location, self.version, self.oldTag, old.newContent, old.conflict )
      elif isinstance( old, AddBefore ):
        # new must add before the element following the deleted element
        newLocation = self.location[:-1] + ( self.location[-1]+1 ,)
        return AddBefore( newLocation, self.version, old.newTag, old.newContent, old.conflict )
    elif old.location < self.location:
      # new is unaffected by reordering as before deleted element
      return old
    elif old.location[:-1] != self.location[:-1]:
      # new is unaffected by reordering as does not share parent with deleted element
      return old
    else:
      # new's location moves up one place
      old.location = old.location[:-1] + ( old.location[-1]-1 ,)
      return old

  def apply( self, tree ):
    # FIXME: support conflicts
    x=tree.tree
    for index in self.location[1:-1]:
      x = x[index]

    child=x[self.location[-1]]

    if child.text!=self.oldContent and child.text is not None:
      # FIXME: we allow editors to specify the wrong "old" value with a warning
      logger.log(1, 'expected content "',self.oldContent,'", but found "',child.text,'"')
      self.oldContent = child.text

    if child.tag!=self.oldTag:
      # FIXME: we allow editors to specify the wrong "old" value with a warning
      logger.log(1, 'expected tag ',self.oldTag,', but found ',child.tag)
      self.oldTag = child.tag

    x.remove(child)

  def __repr__( self ):
    return '<Delete '+self.oldTag+' "'+self.oldContent+'" at '+str(self.location)+'>'



class AddBefore(Delta):
  '''Add the tag and content specified before location.

location may be "off the end" of the current tree in which case the
element is appended.
'''

  def __init__( self, location, version, newTag, newContent, conflict=[] ):
    Delta.__init__(self, location, version, conflict)
    self.newTag = newTag
    self.newContent = newContent

  def swap( self, old ):
    if old.location < self.location:
      # new is unaffected by reordering as before inserted element
      return old
    elif old.location[:-1] != self.location[:-1]:
      # new is unaffected by reordering as does not share parent with inserted element
      return old
    else:
      # new's location moves down one place
      logger.log( 3, 'Location', old.location )
      logger.log( 3, 'Location', old.location[:-1] )
      logger.log( 3, 'Location', old.location[-1]+1 )
      old.location = old.location[:-1] + ( old.location[-1]+1 ,)
      return old

  def apply( self, tree ):
    # FIXME: support conflicts

    x=tree.tree
    for index in self.location[1:-1]:
      x = x[index]

    newTree = pad.tree.SimpleTree( '<'+self.newTag+'>'+self.newContent+'</'+self.newTag+'>' )
    x.insert( self.location[-1], newTree.tree )

  def __repr__( self ):
    return '<Add '+self.newTag+' "'+self.newContent+'" before '+str(self.location)+'>'



class Replace(Delta):
  '''Replace the content of an element.

This delta should only be used where the schema requires that an element be present.
In other cases, use Delete then AddBefore.
'''

  def __init__( self, location, version, oldContent, newContent, conflict=[] ):
    if oldContent is None: oldContent=''
    if newContent is None: newContent=''

    Delta.__init__(self, location, version, conflict)
    self.oldContent = oldContent
    self.newContent = newContent

  def swap( self, old ):
    if old.location == self.location:
      if isinstance( old, Replace ):
        # new must replace my newContent instead of old's oldContent - causes a conflict
        return Replace( self.location, self.version, self.newContent, old.newContent, old.conflict + [self.newContent] )
      elif isinstance( old, Delete ):
        # new must delete my newContent instead of old's oldContent - causes a conflict
        return Delete( self.location, self.version, old.oldTag, self.newContent, old.conflict + [self.newContent] )
      elif isinstance( old, AddBefore ):
        # replacement does not affect the rest of the tree
        return old
    else:
      # replacement does not affect the rest of the tree
      return old

  def apply( self, tree ):
    x=tree.tree
    logger.log(3,'using location ',self.location)
    for index in self.location[:-1]:
      x = x[index]
      logger.log(3,'entered element ',x.tag,' at position ',index)

    if isinstance(self.location[-1],int):
      # content is treated as an atomic string - even if it actually includes child elements
      # because a delta works with the "unit of editability"

      if x.text!=self.oldContent and x.text is not None:
        # FIXME: we allow editors to specify the wrong "old" value with a warning
        logger.log(1,'expected content "',self.oldContent,'", but found "',x.text,'"')
        self.oldContent = x.text

      oldElement = x[self.location[-1]]

      if self.conflict:
        newContent = ''
        for item in self.conflict:
          newContent += '<'+oldElement.tag+'>'+item+'</'+oldElement.tag+'>'
        newContent += '<'+oldElement.tag+'>'+self.newContent+'</'+oldElement.tag+'>'
        newTree = pad.tree.SimpleTree( '<pad:conflict xmlns:pad="http://www.colourcountry.net/xmlns/pad">'+newContent+'</pad:conflict>' )
      else:
        newTree = pad.tree.SimpleTree( '<'+oldElement.tag+'>'+self.newContent+'</'+oldElement.tag+'>' )
      x.remove( oldElement )
      x.insert( self.location[-1], newTree.tree )
      
    elif self.location[-1].startswith("@"):
      attributeName = self.location[-1][1:]
      text = x.get( attributeName )
      logger.log(3,'got content ',text,' from attribute ',attributeName )
      if text!=self.oldContent:
        # FIXME: we allow editors to specify the wrong "old" value with a warning
        logger.log(1,'expected content "',self.oldContent,'", but found "',text,'"')
        self.oldContent = text

      x.set(self.location[-1][1:],self.newContent)
    else:
      raise ValueError("Badly formed location")

  def __repr__( self ):
    return '<Replace "'+self.oldContent+'" with "'+self.newContent+'" at '+str(self.location)+'>'





class DeltaList:
  def __init__( self ):
    self.deltas = {}

  def append( self, delta, newVersion ):
    '''Swaps a delta to the end of the delta list, then appends, with the new version number.

@param delta The delta to add to the list.
@param newVersion The new version number. This must be greater than any delta in the list.
@return The delta which was finally appended.
'''

    newVersion = unicode(newVersion)
    keys = filter( lambda x:x>=delta.version, self.deltas.keys() )

    logger.log( 3,'Versions later than ',delta.version,': ',keys )

    if keys:
      keys.sort()
      for laterVersion in keys:
        logger.log(3, 'Moving past delta  ', self.deltas[laterVersion], '(',laterVersion,')' )
        logger.log(3, 'Swapping out delta ', delta)
        delta = self.deltas[laterVersion].swap( delta )
        logger.log(3, 'Swapping in delta  ', delta)
      if keys[-1]>newVersion:
        raise ValueError("Supplied new version ("+str(newVersion)+") was lower than the latest delta ("+str(keys[-1])+")")
    else:
      if delta.version>newVersion:
        raise ValueError("Supplied new version ("+str(newVersion)+") was lower than the delta to be applied ("+str(keys[-1])+")")

    delta.version = newVersion
    self.deltas[delta.version] = delta

    return delta



class DeltaStorage:
  def __init__( self, path ):

    # Give pickler access to the defined classes
    sys.modules['__main__'].DeltaList = DeltaList
    sys.modules['__main__'].Delta = Delta
    sys.modules['__main__'].Replace = Replace
    sys.modules['__main__'].Delete = Delete
    sys.modules['__main__'].AddBefore = AddBefore

    self.path = os.path.join( path, 'current' )
    self.delta = os.path.join( path, 'delta' )
    self.magic = {}
    try:
      os.makedirs( self.path )
    except OSError:
      pass # probably already exists.
    try:
      os.makedirs( self.delta )
    except OSError:
      pass # probably already exists.

  def _getNsPath( self, ns, path=None ):
    if not path:
      return os.path.join( self.path, nsquote(ns) )
    if path.endswith('/'):
      path=path[:-1]

    bits = [ self.path, nsquote(ns) ] + [ filenamequote(bit) for bit in path.split('/') ]
    return os.path.join( *bits[:-1] )


  def _getFullPath( self, ns, path ):
    if path.endswith('/'):
      path=path[:-1]

    bits = [ self.path, nsquote(ns) ] + [ filenamequote(bit) for bit in path.split('/') ]
    return os.path.join( *bits )


  def _read( self, ns, path ):

    content = None
    isReadOnly = False

    for magicPath in self.magic:
      if path==magicPath or path.startswith(magicPath+'/'):
        logger.log( 3, path, " matched magic path ",magicPath )
        content = self.magic[magicPath]( self, ns, path, *path[len(magicPath)+1:].split('/') )
        isReadOnly = True

    if not content:
      fullPath = self._getFullPath( ns, path )
      try:
        content = file( fullPath, 'r' ).read()
        logger.log( 3, 'Read ',fullPath )
      except IOError: # didn't exist
        logger.log( 3, "Couldn't read ",fullPath )
        pass

    return content, isReadOnly

  def _write( self, ns, path, tree ):
    nsPath = self._getNsPath( ns )
    fullPath = self._getFullPath( ns, path )
    logger.log( 3, 'Writing: ',fullPath )
    try:
      os.makedirs( nsPath )
    except OSError:
      pass
    file( fullPath, 'w' ).write( tree.getXML() )
    return False # success

  def _readDeltas( self, ns, path ):
    nsPath = os.path.join( self.delta, filenamequote(ns) )
    fullPath = os.path.join( nsPath, filenamequote(path)+'.delta' )
    logger.log( 3, 'Reading deltas: ',fullPath )
    try:
      return pickle.load( file( fullPath, 'r' ) )
    except IOError: # didn't exist, make it
      return DeltaList()

  def _writeDeltas( self, ns, path, deltas ):
    nsPath = os.path.join( self.delta, filenamequote(ns) )
    fullPath = os.path.join( nsPath, filenamequote(path)+'.delta' )
    logger.log( 3, 'Writing deltas: ',fullPath )
    try:
      os.makedirs( nsPath )
    except OSError:
      pass
    pickle.dump( deltas, file( fullPath, 'w' ) )
    return None

  def get( self, ns, path ):
    content, isReadOnly = self._read( ns, path )
    if not content:
      return None

    return pad.tree.Tree( content, time.time(), ns, path, isReadOnly )

  def apply( self, ns, path, delta, fallbackFn=None ):
    if delta is None: return

    tree = self.get( ns, path )

    if tree is None:
      if fallbackFn is None:
        logger.log(ValueError,'You must supply a function which will generate a fallback tree')
      tree = fallbackFn()

    deltas = self._readDeltas( ns, path )
    logger.log( 3, 'Request to apply delta ',delta,' at version ',delta.version )

    delta = deltas.append( delta, time.time() )
    logger.log( 3, 'Applying delta ',delta )

    delta.apply( tree )
    if delta.conflict:
      # don't save this tree as probably not consistent with content model
      return tree
    else:
      self._write( ns, path, tree )
      self._writeDeltas( ns, path, deltas )
      return None

  def addMagic(self, path, func):
    if path.endswith('/'):
      path = path[:-1]
    self.magic[path] = func

  def listPrecedingSiblings( self, ns, path, number ):
    '''List the <number> paths whose names are next in reverse alphabetical order from the given path.

Exception: Passing '' or a bare directory name ending '/' yields the <number> paths which are first in reverse alphabetical order.'''

    nsPath = self._getNsPath( ns, path )

    if path.endswith('/'):
      path=path[:-1]

    try:
      rindex = path.rindex('/')
    except ValueError:
      rindex = None

    existingPaths = [ path[:rindex]+'/'+filenameunquote(x) for x in os.listdir( nsPath ) ]

    if path=='':
      matches = existingPaths
    else:
      matches = filter( lambda x:x<path, existingPaths)
      logger.log(3, 'Adding ',path,' to ',matches)
      matches.append( path )
    matches.sort()
    matches.reverse()
    return [ match for match in matches[:number] ]

  # Helper method to enable views to access the different delta types.

  def getDelta( self, action, location, version, a, b):
    assert isinstance(a, unicode)
    assert isinstance(b, unicode)
    
    if action=='replace':
      # FIXME: we allow replace even if a==b because the editor may get a wrong
      return Replace( location, version, a, b )
    elif action=='delete':
      return Delete( location, version, a, b )
    elif action=='addbefore':
      return AddBefore( location, version, a, b )



def testStorage():
  storage = DeltaStorage( 'content' )

  tree = storage.get( 'http://malachi.colourcountry.net/ns', 'index' )

  deltb = tree.getReplace( (0,3,4), 'Second delta '+tree.version )
  logger.log(3, 'Second delta:',deltb)
  storage.apply( 'http://malachi.colourcountry.net/ns', 'index', deltb )

  tree = storage.get( 'http://malachi.colourcountry.net/ns', 'index' )

  deltc = tree.getAddBefore( (0,3,2), 'p', 'Added back in' )
  deltd = tree.getReplace( (0,3,3), 'Second delta <b>with extra elements</b> in it' )
  storage.apply( 'http://malachi.colourcountry.net/ns', 'index', deltd )
  storage.apply( 'http://malachi.colourcountry.net/ns', 'index', deltc )

  tree = storage.get( 'http://malachi.colourcountry.net/ns', 'index' )
  print tree

  deltd = tree.getReplace( (0,3,3), 'Third delta <i>with new elements</i> in it' )
  storage.apply( 'http://malachi.colourcountry.net/ns', 'index', deltd )

  tree = storage.get( 'http://malachi.colourcountry.net/ns', 'index' )
  print tree
