#!/usr/bin/python

import math

class Element:
    def __init__(self, x, y, scale=1):
        self.x = x
        self.y = y
        self.scale = scale

    def writeSelf(self):
        return ''

    def writeUnscaled(self):
        return self.writeSelf()

    def write(self):
        s = ''
        if self.x != 0 or self.y != 0:
            s += '''
    <g transform="translate(%f,%f)">''' % (self.x, self.y)
        if self.scale != 1:
            s += '''
    <g transform="scale(%f)">''' % self.scale

        s += self.writeUnscaled()

        if self.scale != 1:
            s += '''
    </g>'''
        if self.x != 0 or self.y != 0:
            s += '''
    </g>'''

        return s

class ContainerElement(Element):
    def __init__(self, x, y, scale=1):
        Element.__init__(self, x, y, scale)
        self.underElements = []
        self.overElements = []

    def add(self, element):
        assert isinstance(element, Element)
        self.overElements.append(element)

    def addUnder(self, element):
        assert isinstance(element, Element)
        self.underElements.append(element)

    def writeUnscaled(self):
        s = ''
        for element in self.underElements:
            s += element.write()
        s += self.writeSelf()
        for element in self.overElements:
            s += element.write()
        return s


class RectangularGrid(ContainerElement):
    def __init__(self, width, height, x, y, scale=100):
        ContainerElement.__init__(self, x+scale/2, y+scale/2, scale)
        self.width = width
        self.height = height

    def copyTo(self, x, y, scale=None):
        if scale is None:
            scale = self.scale
        return RectangularGrid(self.width, self.height, x, y, scale)

    def writeSelf(self):
        s = '''
        <g style="stroke:#000000;stroke-width:%f;fill:none">''' % (3.0/self.scale)
        for ix in xrange(0,self.width, 2):
            s += '''
            <rect x="%f" y="-0.5" width="1" height="%f" />''' % (ix-0.5, self.height)

        for iy in xrange(0,self.height, 2):
            s += '''
            <rect x="-0.5" y="%f" width="%f" height="1" />''' % (iy-0.5, self.width)

        s += '''
            <rect x="-0.5" y="-0.5" style="stroke-width:%f"
                  width="%f" height="%f" />''' % (6.0/self.scale, self.width, self.height)

        s += '''
        </g>'''
        return s

class Shade(Element):
    def writeSelf(self):
        return '''
            <rect x="-0.5" y="-0.5"
                  width="1" height="1"
                  stroke="none" fill="#eeeeee" />'''

class Arrow(Element):
    def __init__(self, x, y, tox, toy, scale=1):
        Element.__init__(self, x, y, scale)
        self.tox = tox - x
        self.toy = toy - y

    def writeSelf(self):

        angle = math.atan2( self.tox, self.toy )
        rad45 = 45.0 / 180.0 * math.pi
        alx = self.tox+0.25*math.cos(angle+rad45)
        aly = self.toy-0.25*math.sin(angle+rad45)
        arx = self.tox-0.25*math.cos(angle-rad45)
        ary = self.toy+0.25*math.sin(angle-rad45)

        return'''
            <g style="stroke:#000000;stroke-width:0.03;fill:none">
                <path d="M 0 0 L %f %f M %f %f L %f %f L %f %f"/>''' % (
                    self.tox, self.toy,
                    alx, aly, self.tox, self.toy, arx, ary) + '''
            </g>'''
        

class Piece(Element):
    def __init__(self, name, x, y, scale=1):
        Element.__init__(self, x, y, scale)
        self.name = name
        self.scale = scale

    def writeSelf(self):
        s = '''
            <use xlink:href="pieces-stack.svg#'''+self.name+'''" />'''

        return s

class IceStack(ContainerElement):
    def __init__(self, pieces, x, y):
        ContainerElement.__init__(self, x, y, 1)
        myy = 0.26
        for piece in pieces:
            piece.x = 0
            piece.y = myy
            self.add( piece )
            myy -=0.26

class IcePiece(Piece):
    SIZES = {
        0: "small",
        1: "medium",
        2: "large",
        "small": "small",
        "medium": "medium",
        "large": "large",
    }

    COLOURS = {
        "black": "9",
        "purple": "8",
        "blue": "7",
        "teal": "6",
        "red": "5",
        "orange": "4",
        "green": "3",
        "clear": "2",
        "yellow": "1",
        "white": "0",
    }

    def __init__(self, colour, size, x=0, y=0):
        name = IcePiece.SIZES[size] + IcePiece.COLOURS[colour]
        Piece.__init__(self, name, x, y, 0.005)


class SvgLayout(ContainerElement):
    def __init__(self):
        ContainerElement.__init__(self, 0, 0, 1)
        self.width = 4200
        self.height = 5940
        self.elements = []

    def write(self):
        s = ''
        s += '''<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<svg
   xmlns:svg="http://www.w3.org/2000/svg"
   xmlns:xlink="http://www.w3.org/1999/xlink"
   xmlns="http://www.w3.org/2000/svg"
   version="1.0"
   width="'''+str(self.width)+'''"
   height="'''+str(self.height)+'''">'''

        s += self.writeUnscaled()

        s += '''
</svg>
'''
        return s



l = SvgLayout()

board = RectangularGrid(5,5,0,0)

grid = board.copyTo(200,200)

pieces = [  IcePiece("blue",1),
            IcePiece("red",2),
            IcePiece("green",2),
            IcePiece("white",0),
            IcePiece("white",0) ]

grid.add(IceStack(pieces, 2, 2))
grid.add(IcePiece("white",0, 3, 3))
grid.addUnder(Shade(0,1))
grid.addUnder(Shade(1,0))
grid.addUnder(Shade(0,3))
grid.addUnder(Shade(3,0))
grid.add(Arrow(2,2,0,1))
l.add(grid)


print l.write()
